import { MessageTransmissionTransmitter } from "./msg.transmission.transmitter";
import { MessageTransmissionReceiver } from "./msg.transmission.receiver";
import { AdapterManager } from "../connector/adapter.manager";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionRequestResponse } from "./msg.transmission.request-response";
import { filter, Observable, Observer, Subject } from "rxjs";
import ConsoleLogger from "../utils/log.utils";
import { TransmitterAdapter } from "../connector/adapter.transmitter";
import { ReceiverAdapter } from "../connector/adapter.receiver";
import { AdapterEvent, AdapterInterface, EventMessage, EventObject, GeneralEvent, Transmission, TransmissionEvent, TransportEvent, TransportService } from '../interface/interface'
import { MessageTransmissionManagerBase } from "./msg.transmission.manager.base";

export class MessageTransmissionManager extends MessageTransmissionManagerBase {
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionManager`, ['managers'])
    private browserEnv!: boolean
    private transmissionSet: Transmission[] = []
    private adapterManager!: AdapterManager

    constructor(transportRef: Observable<TransportService>, browserEnv?: boolean) {
        super()
        this.console.log({ message: `Constructing self...` })
        this.eventObj.globalEvent = new Subject()
        this.eventObj.transportEvent = new Subject()
        transportRef.subscribe((transport: TransportService) => {
            this.eventObj.transportEvent.next({
                id: uuidv4(),
                type: `Transport Event`,
                event: `New Transport`,
                date: new Date(),
                data: transport
            } as TransportEvent)
        })
        if (browserEnv) this.browserEnv = browserEnv
        // Subscribe for adapterManager and it's relevent event
        this.adapterManager = new AdapterManager(this.eventObj, browserEnv)
        this.adapterManager.subscribe().subscribe(this.eventObj.adapterEvent)

    }

    public subscribe(): Observable<TransmissionEvent> {
        return new Observable((observer: Observer<TransmissionEvent>) => {
            const targetEvent: EventMessage = this.browserEnv ? 'New Server' : 'New Client';
            this.eventObj.transportEvent.pipe(
                filter(event => event.event == targetEvent)
            ).subscribe(event => {
                // get all adapters for all the connection
                let transmission: Transmission | undefined = this.instantiateTransmissionComponents(event?.data?.clientId)
                if (transmission) {
                    observer.next({
                        id: uuidv4(),
                        type: `Transmission Event`,
                        event: 'New Transmission',
                        date: new Date(),
                        transmission: transmission
                    })
                }
            })
        })
    }

    private instantiateTransmissionComponents(clientId: string): Transmission {
        let receiverInstance: MessageTransmissionReceiver = this.getReceiver(clientId, this.eventObj)
        let transmitterInstance: MessageTransmissionTransmitter = this.getTransmitter(clientId, this.eventObj)
        let requestResponseInstance: MessageTransmissionRequestResponse = this.getRequestResponse(clientId, this.eventObj, transmitterInstance, receiverInstance)
        let transmissionObj: Transmission = {
            clientId: clientId,
            transmitter: transmitterInstance,
            receiver: receiverInstance,
            requestResponse: requestResponseInstance,
            event: this.eventObj.globalEvent
        }

        return transmissionObj
    }

    private getReceiver(clientId: string, eventObj: EventObject): MessageTransmissionReceiver {
        throw new Error(`Method not defined`)
    }
    private getTransmitter(clientId: string, eventObj: EventObject): MessageTransmissionTransmitter {
        throw new Error(`Method not defined`)
    }
    private getRequestResponse(clientId: string, eventObj: EventObject, transmitter: MessageTransmissionTransmitter, receiver: MessageTransmissionReceiver): MessageTransmissionRequestResponse {
        throw new Error(`Method not defined`)
    }


}


