import { MessageTransmissionBase } from "./msg.transmission.base";
import { v4 as uuidv4 } from 'uuid'
import { BehaviorSubject, distinct, distinctUntilChanged, filter, map, Observable, Subject } from "rxjs";
import { RetransmissionService } from "../utils/retransmission.service";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { Adapter } from "../connector/adapter.base";
import { TransmitterAdapter } from "../connector/adapter.transmitter";
import { AdapterEvent, AdapterInterface, ConnectionState, EventObject, FisMessage, MessageTransmitterInterface, TransportEvent, TransportMessage } from "../interface/interface";

/* Take in all the messages that needs to be transported, and divide them accordingly. So the connector instances will do just that
connectors or adapters will have their own identifier*/
export class MessageTransmissionTransmitter extends MessageTransmissionBase implements MessageTransmitterInterface {
    private connectionStateEvent: BehaviorSubject<ConnectionState> = new BehaviorSubject<ConnectionState>('OFFLINE')
    private console: ConsoleLogger = new ConsoleLogger(`MessageTransmissionTransmitter`, ['transmission'])
    private messageToBeTransmitted!: Subject<FisMessage | WrappedMessage>
    private buffer!: RetransmissionService;
    private currentAdapter!: TransmitterAdapter

    constructor(clientId: string, eventObj: EventObject) {
        super()
        this.console.log({ message: `Constructing Transmitter Transmission with ${clientId}` })
        this.eventObj = eventObj
        this.messageToBeTransmitted = new Subject()
        this.buffer = new RetransmissionService()
        this.setupBuffer()
        this.handleAdapters(eventObj.adapterEvent)

        // special case just for http in case of server/client disconnected, the unsent msg will be flushed back into messageToBeTransmitted
        // logic here
    }

    public emit(message: FisMessage): void {
        this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting message` : `Buffering message`}` })
        this.messageToBeTransmitted.next(message)
    }

    private setupBuffer(): void {
        this.console.log({ message: `Setting up Retransmission Service...` })
        this.eventObj.transportEvent.pipe(
            filter(event => event.data.clientId == this.clientId),
            filter(event => event.event == 'Client Disconnected' || event.event == 'Client Re-connected' || event.event == 'Client Connected' || event.event == 'Server Disconnected' || event.event == 'Server Connected'),
            map(event => {
                if (event.event == 'Client Disconnected' || event.event == 'Server Disconnected') {
                    return 'OFFLINE'
                } else {
                    return `ONLINE`
                }
            }),
            distinctUntilChanged()
        ).subscribe((signal: ConnectionState) => {
            this.connectionStateEvent.next(signal)
            if (signal == 'OFFLINE') this.console.error({ message: `${this.clientId} disconnected` })
            if (signal == 'ONLINE') this.console.log({ message: `${this.clientId} connected` })
        })
        this.buffer.implementRetransmission(this.messageToBeTransmitted, this.connectionStateEvent.asObservable(), true)
        // automatically subscribe to allow released bffered messages to be released
        this.buffer.returnSubjectForBufferedItems().subscribe((bufferedMessage: WrappedMessage) => {
            // need to work with wrapped messages
            this.console.log({ message: `Releasing ${bufferedMessage.thisMessageID}` });
            (this.currentAdapter as TransmitterAdapter).emit(bufferedMessage)
        })
    }

    private handleAdapters(adaptersEvent: Observable<AdapterEvent>): void {
        if (Array.isArray(event.adapters)) {
            if (event.adapters.length > 0) {
                event.adapters.forEach(adapter => {
                    this.adapters.push(adapter)
                })
            }
        } else {
            this.adapters.push(event.adapters)
        }
    }

    private handleNewAdapters(adaptersEvent: Observable<AdapterEvent>): void {
        adaptersEvent.pipe(
            filter(event => event.event === `New Adapter`),
            map(event => { return event.adapters }),
        ).subscribe({
            next: (adapters: ) => {

            }
        })
    }

    private handleAdaptersTermination(adaptersEvent: Observable<AdapterEvent>): void {
        adaptersEvent.subscribe
    }

    private setUpAdapter(adapter: TransmitterAdapter): void {
        this.currentAdapter = adapter
    }

    private uniqueHandlerToFlushUnsentMessages(event: Observable<TransportEvent>): void {
        event.pipe(
            filter(event => event.event == 'Re-Flush'),
            filter(event => event.data.clientId == this.clientId),
        ).subscribe((event: TransportEvent) => {
            this.console.log({ message: `${this.connectionStateEvent.getValue() == 'ONLINE' ? `Transmitting ${((event.data.payload as TransportMessage).payload as WrappedMessage).thisMessageID}` : `Buffering ${((event.data.payload as TransportMessage).payload as WrappedMessage).thisMessageID}`}` })
            this.messageToBeTransmitted.next(((event.data.payload as TransportMessage).payload as WrappedMessage))
        })
    }

}